#version 460 core

#ifndef IS_IRIS
	#error This shader pack requires Iris: https://irisshaders.dev/
#endif

#include "/lib/settings.glsl"

/*
	const float ambientOcclusionLevel = 0.8; // [0.0 0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0]

	const int colortex0Format = RGB16F;
*/

/* RENDERTARGETS: 0 */
layout(location = 0) out vec3 outColor0;

uniform float viewHeight, viewWidth;
uniform sampler2D colortex0, colortex1;

in vec2 texCoord;

#if BLOOM
	const int noiseTextureResolution = 256;

	uniform sampler2D noisetex;
#endif

#ifdef AUTOMATIC_EXPOSURE
	/*
		const float eyeBrightnessHalflife = 10.0; // [0.0 0.5 1.0 1.5 2.0 2.5 3.0 3.5 4.0 4.5 5.0 5.5 6.0 6.5 7.0 7.5 8.0 8.5 9.0 9.5 10.0]
	*/

	uniform ivec2 eyeBrightnessSmooth;

	uniform int worldTime;
#endif

#if TONEMAP == 8 // Uncharted 2
	vec3 uncharted2Math(in vec3 color) {
		const float A = 0.15;
		const float B = 0.5;
		const float C = 0.1;
		const float D = 0.2;
		const float E = 0.02;
		const float F = 0.3;
		const float W = 11.2;

		return ((color * (A * color + C * B) + D * E) / (color * (A * color + B) + D * F)) - E / F;
	}
#endif

vec3 tonemap(in vec3 color) {
	#if !TONEMAP
		return color;
	#elif TONEMAP == 1 // Uchimura
		const float P = 1.0; // Max display brightness
		const float a = 1.0; // Contrast
		const float m = 0.22; // Linear section start
		const float l = 0.4; // Linear section length
		const float c = 1.33; // Black
		const float b = 0.0; // Pedestal

		const float l0 = (P - m) * l / a;
		const float L0 = m - m / a;
		const float L1 = m + (1.0 - m) / a;
		const float S0 = m + l0;
		const float S1 = m + a * l0;
		const float C2 = a * P / (P - S1);
		const float CP = -C2 / P;

		vec3 w0 = 1.0 - smoothstep(0.0, m, color);
		vec3 w2 = step(m + l0, color);
		vec3 w1 = 1.0 - w0 - w2;

		vec3 T = m * pow(color / m, vec3(c)) + b;
		vec3 S = P - (P - S1) * exp(CP * (color - S0));
		vec3 L = m + a * (color - m);

		return T * w0 + L * w1 + S * w2;
	#elif TONEMAP == 2 // Reinhard-Jodie
		color = pow(color, vec3(TONEMAP_CURVE));

		float luminance = dot(color, vec3(0.2125, 0.7154, 0.0721)) + 1.0;

		vec3 tv = color / (1.0 + color);

		return pow(mix(color / luminance, tv, tv), vec3(1.0 / TONEMAP_CURVE));
	#elif TONEMAP == 3 // ACES
		const float a = 2.51;
		const float b = 0.03;
		const float c = 2.43;
		const float d = 0.59;
		const float e = 0.14;

		return clamp((color * (a * color + b)) / (color * (c * color + d) + e), 0.0, 1.0);
	#elif TONEMAP == 4 // Lottes
		const vec3 hdrMax = vec3(8.0);
		const vec3 midIn = vec3(0.18);
		const vec3 midOut = vec3(0.267);

		const vec3 a = vec3(1.6);
		const vec3 d = a * vec3(0.977);

		const vec3 b = (-pow(midIn, a) + pow(hdrMax, a) * midOut) / ((pow(hdrMax, d) - pow(midIn, d)) * midOut);
		const vec3 c = (pow(hdrMax, d) * pow(midIn, a) - pow(hdrMax, a) * pow(midIn, d) * midOut) / ((pow(hdrMax, d) - pow(midIn, d)) * midOut);

		return pow(color, a) / (pow(color, d) * b + c);
	#elif TONEMAP == 5 // Unreal
		return pow(color / (color + 0.155) * 1.019, vec3(2.2));
	#elif TONEMAP == 6 // Filmic
		color = max(vec3(0.0), color - 0.004);

		return pow(color * (6.2 * color + 0.5) / (color * (6.2 * color + 1.7) + 0.06), vec3(2.2));
	#elif TONEMAP == 7 // Burgess HDR
		vec3 overExposed = color * 1.2;
		vec3 underExposed = color * 0.6;

		color = mix(underExposed, overExposed, color);

		color = max(vec3(0.0), color - 0.004);

		return color * (6.2 * color + 0.05) / (color * (6.2 * color + 2.3) + 0.06);
	#elif TONEMAP == 8 // Uncharted 2
		const float W = 11.2;
		const float exposureBias = 2.0;
		
		return uncharted2Math(color * exposureBias) / uncharted2Math(vec3(W));
	#endif
}

#ifdef FXAA
	float luminance(in vec3 color) {
		return dot(color, vec3(0.299, 0.587, 0.114));
	}
#endif

void main() {
	vec3 color = texture(colortex0, texCoord).rgb;

	#ifdef FXAA
		// FXAA code from BSL Shaders by Capt Tatsu (https://bitslablab.com)
		// FXAA 3.11 from http://blog.simonrodriguez.fr/articles/30-07-2016_implementing_fxaa.html

		const float edgeThresholdMin = 0.03125;
		const float edgeThresholdMax = 0.125;

		vec2 view = 1.0 / vec2(viewWidth, viewHeight);

		float lumaCenter = luminance(color);
		float lumaDown = luminance(texture(colortex0, texCoord + vec2(0.0, -1.0) * view).rgb);
		float lumaUp = luminance(texture(colortex0, texCoord + vec2(0.0, 1.0) * view).rgb);
		float lumaLeft = luminance(texture(colortex0, texCoord + vec2(-1.0, 0.0) * view).rgb);
		float lumaRight = luminance(texture(colortex0, texCoord + vec2(1.0, 0.0) * view).rgb);

		float lumaMin = min(lumaCenter, min(min(lumaDown, lumaUp), min(lumaLeft, lumaRight)));
		float lumaMax = max(lumaCenter, max(max(lumaDown, lumaUp), max(lumaLeft, lumaRight)));

		float lumaRange = lumaMax - lumaMin;

		if (lumaRange > max(edgeThresholdMin, lumaMax * edgeThresholdMax)) {
			const float quality[12] = float[12] (
				1.0,
				1.0,
				1.0,
				1.0,
				1.0,
				1.5,
				2.0,
				2.0,
				2.0,
				2.0,
				4.0,
				8.0
			);

			const float subpixelQuality = 0.75;

			float lumaDownLeft = luminance(texture(colortex0, texCoord + vec2(-1.0, -1.0) * view).rgb);
			float lumaUpRight = luminance(texture(colortex0, texCoord + vec2(1.0, 1.0) * view).rgb);
			float lumaUpLeft = luminance(texture(colortex0, texCoord + vec2(-1.0, 1.0) * view).rgb);
			float lumaDownRight = luminance(texture(colortex0, texCoord + vec2(1.0, -1.0) * view).rgb);

			float lumaDownUp = lumaDown + lumaUp;
			float lumaLeftRight = lumaLeft + lumaRight;

			float lumaLeftCorners = lumaDownLeft + lumaUpLeft;
			float lumaDownCorners = lumaDownLeft + lumaDownRight;
			float lumaRightCorners = lumaDownRight + lumaUpRight;
			float lumaUpCorners = lumaUpRight + lumaUpLeft;

			float edgeHorizontal =
				abs(-2.0 * lumaLeft + lumaLeftCorners) +
				abs(-2.0 * lumaCenter + lumaDownUp) * 2.0 +
				abs(-2.0 * lumaRight + lumaRightCorners);
			float edgeVertical =
				abs(-2.0 * lumaUp + lumaUpCorners) +
				abs(-2.0 * lumaCenter + lumaLeftRight) * 2.0 +
				abs(-2.0 * lumaDown + lumaDownCorners);

			bool isHorizontal = (edgeHorizontal >= edgeVertical);

			float luma1 = isHorizontal ? lumaDown : lumaLeft;
			float luma2 = isHorizontal ? lumaUp : lumaRight;
			float gradient1 = luma1 - lumaCenter;
			float gradient2 = luma2 - lumaCenter;

			bool is1Steepest = abs(gradient1) >= abs(gradient2);
			float gradientScaled = 0.25 * max(abs(gradient1), abs(gradient2));

			float stepLength = isHorizontal ? view.y : view.x;

			float lumaLocalAverage = 0.0;

			if (is1Steepest) {
				stepLength = - stepLength;
				lumaLocalAverage = 0.5 * (luma1 + lumaCenter);
			} else {
				lumaLocalAverage = 0.5 * (luma2 + lumaCenter);
			}

			vec2 currentUv = texCoord;

			if (isHorizontal) {
				currentUv.y += stepLength * 0.5;
			} else {
				currentUv.x += stepLength * 0.5;
			}

			vec2 offset = isHorizontal ? vec2(view.x, 0.0) : vec2(0.0, view.y);

			vec2 uv1 = currentUv - offset;
			vec2 uv2 = currentUv + offset;

			float lumaEnd1 = luminance(texture(colortex0, uv1).rgb) - lumaLocalAverage;
			float lumaEnd2 = luminance(texture(colortex0, uv2).rgb) - lumaLocalAverage;

			bool reached1 = abs(lumaEnd1) >= gradientScaled;
			bool reached2 = abs(lumaEnd2) >= gradientScaled;

			bool reachedBoth = reached1 && reached2;

			if (!reached1) uv1 -= offset;
			if (!reached2) uv2 += offset;

			if (!reachedBoth) {
				for(uint i = 2; i < 12; i++) {
					if (!reached1) lumaEnd1 = luminance(texture(colortex0, uv1).rgb) - lumaLocalAverage;
					if (!reached2) lumaEnd2 = luminance(texture(colortex0, uv2).rgb) - lumaLocalAverage;

					reached1 = abs(lumaEnd1) >= gradientScaled;
					reached2 = abs(lumaEnd2) >= gradientScaled;
					reachedBoth = reached1 && reached2;

					if (!reached1) uv1 -= offset * quality[i];
					if (!reached2) uv2 += offset * quality[i];

					if (reachedBoth) break;
				}
			}

			float distance1 = isHorizontal ? (texCoord.x - uv1.x) : (texCoord.y - uv1.y);
			float distance2 = isHorizontal ? (uv2.x - texCoord.x) : (uv2.y - texCoord.y);

			bool isDirection1 = distance1 < distance2;
			float distanceFinal = min(distance1, distance2);

			float edgeThickness = (distance1 + distance2);

			float pixelOffset = - distanceFinal / edgeThickness + 0.5;

			bool isLumaCenterSmaller = lumaCenter < lumaLocalAverage;

			bool correctVariation = ((isDirection1 ? lumaEnd1 : lumaEnd2) < 0.0) != isLumaCenterSmaller;

			float finalOffset = correctVariation ? pixelOffset : 0.0;

			float lumaAverage = (1.0 / 12.0) * (2.0 * (lumaDownUp + lumaLeftRight) + lumaLeftCorners + lumaRightCorners);
			float subPixelOffset1 = clamp(abs(lumaAverage - lumaCenter) / lumaRange, 0.0, 1.0);
			float subPixelOffset2 = (-2.0 * subPixelOffset1 + 3.0) * subPixelOffset1 * subPixelOffset1;
			float subPixelOffsetFinal = subPixelOffset2 * subPixelOffset2 * subpixelQuality;

			finalOffset = max(finalOffset, subPixelOffsetFinal);

			vec2 finalUv = texCoord;

			if (isHorizontal) {
				finalUv.y += finalOffset * stepLength;
			} else {
				finalUv.x += finalOffset * stepLength;
			}

			color = texture(colortex0, finalUv).rgb;
		}
	#endif

	#if BLOOM
		const vec2[9] samplePoints = vec2[9](
			vec2(0.0, 0.0),
			vec2(0.0, 1.0),
			vec2(0.7, 0.7),
			vec2(1.0, 0.0),
			vec2(0.7,-0.7),
			vec2(0.0,-1.0),
			vec2(-0.7,-0.7),
			vec2(-1.0, 0.0),
			vec2(-0.7, 0.7)
		);

		float rotation = texture(noisetex, texCoord * vec2(viewWidth, viewHeight) / noiseTextureResolution).r;

		float cosTheta = cos(rotation);
		float sinTheta = sin(rotation);

		mat2 rotationMatrix = mat2(
			cosTheta, -sinTheta,
			sinTheta, cosTheta
		) * 0.0025;

		vec3 sum = vec3(0.0);

		for(uint i = 0; i < samplePoints.length(); i++) {
			vec2 coord = BLOOM_RADIUS * vec2(samplePoints[i]) * rotationMatrix + texCoord;

			vec3 point = texture(colortex0, coord).rgb;

			sum += point; // * point * point * point * point;
		}

		sum /= samplePoints.length();

		sum *= sum * sum * sum;

		// float brightness = 1.0 + max(color.r, max(color.g, color.b));
		// sum /= brightness;

		color += sum * BLOOM * 0.1;
	#endif

	color = pow(color, vec3(2.4));

	#ifdef AUTOMATIC_EXPOSURE
		float torchLight = eyeBrightnessSmooth.s;
		torchLight /= 30.0;

		float skyLight = eyeBrightnessSmooth.t;
		skyLight /= 240.0;
		skyLight += 0.5 - distance(worldTime, 6000.0) / 12000.0;

		color *= max(1.0, 3.0 - mix(torchLight, 2.0, clamp(skyLight, 0.0, 1.0)));
	#endif

	#if HDR_OVEREXPOSED != 100 || HDR_UNDEREXPOSED != 100
		color = mix(color * HDR_UNDEREXPOSED, color * HDR_OVEREXPOSED, color) * 0.01;
	#endif

	#if RED != 100 || GREEN != 100 || BLUE != 100
		color *= vec3(RED, GREEN, BLUE) * 0.01;
	#endif

	#if SATURATION != 100
		color = mix(dot(color, vec3(0.2126, 0.7152, 0.0722)).rrr, color, SATURATION * 0.01);
	#endif

	// color *= 0.2 * color + 0.9;

	outColor0 = pow(tonemap(color), vec3(1.0 / GAMMA));
}